#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <cassert>
#include <ctime>
#include <cfloat>

// libraries used to convert the chromosome names in yeast
#include <map>
#include <string>
#include <iostream>
using namespace std;

// for random seed, Diament & Tuller, May 2014
#include <sys/types.h>
#include <unistd.h>

#include "genome_model.hpp"
#include "utils.h"
#include "coords.h"
#define NUM_CHROMS 16  

GENOME::GENOME(int bp_per_locus, double min_dist, double max_dist,
		char *interactions_filename, int *chromosome, int add_rDNA,
		char *rDNA_interactions_filename, int use_weights,
		char* unseen_interactions_filename, char *secondary_inter_filename) {
	// added seconday_interactions_filename, Diament & Tuller, May 2014
	this->bp_per_locus = bp_per_locus;
	this->min_locus_distance = min_dist;
	this->max_locus_distance = max_dist;
	this->interactions_filename = interactions_filename;
	this->rDNA_interactions_filename = rDNA_interactions_filename;
	this->wanted_chrom = chromosome;
	this->add_rDNA = add_rDNA;

	// Check to be sure the parameters make sense.
	if (min_locus_distance >= max_locus_distance) {
		die("Minimum distance (%g) >= maximum distance (%g).\n",
				min_locus_distance, max_locus_distance);
	}

	// improved seeding, Diament & Tuller, May 2014
#ifdef __linux__
	// this is a better seed
	FILE *fid = fopen("/dev/urandom", "r");
	long int seed;
	fread( &seed, sizeof(seed), 1, fid);
	fclose(fid);
	printf( "init random seed from /dev/random\nseed:\t%lu\n", seed);
#else
	long int seed = (long int) mix96( clock(), time(NULL), getpid());
	printf( "init random seed from clock, time, getpid\nseed:\t%lu\n", seed);
#endif

	// Initialize the random number generator.
	my_srand(seed);
	num_chroms = 0;
	num_interactions = 0;
	num_rDNA_interactions = 0;
	interactions = NULL;
	rDNA_interactions = NULL;

	// fill up the array of chromosome lengths
	chrlen[1] = 230208;
	chrlen[2] = 813178;
	chrlen[3] = 316617;
	chrlen[4] = 1531918;
	chrlen[5] = 576869;
	chrlen[6] = 270148;
	chrlen[7] = 1090947;
	chrlen[8] = 562643;
	chrlen[9] = 439885;
	chrlen[10] = 745745;
	chrlen[11] = 666454;
	if (add_rDNA)
		chrlen[12] = 1078175 + SIZE_RDNA * NUM_RDNA_TO_ADD;
	else
		chrlen[12] = 1078175;
	chrlen[13] = 924429;
	chrlen[14] = 784333;
	chrlen[15] = 1091289;
	chrlen[16] = 948062;

	// initialize the array of centromere locations
	centromeres[0] = 151584;
	centromeres[1] = 238325;
	centromeres[2] = 114499;
	centromeres[3] = 449819;
	centromeres[4] = 152103;
	centromeres[5] = 148622;
	centromeres[6] = 497042;
	centromeres[7] = 105698;
	centromeres[8] = 355742;
	centromeres[9] = 436418;
	centromeres[10] = 439889;
	centromeres[11] = 150946;
	centromeres[12] = 268149;
	centromeres[13] = 628877;
	centromeres[14] = 326703;
	centromeres[15] = 556070;

	new_haploid_yeast_genome(chromosome);   // (chromosome);
	// handling second interactions file, Diament & Tuller, May 2014
	printf("Adding interactions\n");
	if (use_weights == 5)
		printf("Treating input as Hi-C (use_weights=5)\n");
	else if (use_weights == 6)
		printf("Treating input as a distance function (use_weights=6)\n");

	int inter_priority = 0; // highest
	num_interactions = add_interactions(interactions_filename, interactions,
			chromosome, inter_priority);
	printf("Finished adding main interactions\n");

	if (strcmp(secondary_inter_filename, "") != 0) {
		inter_priority = 1;
		num_interactions += add_interactions(secondary_inter_filename,
				interactions, chromosome, inter_priority);
		printf("Finished adding secondary interactions\n");
	}

	printf("Total interactions: %d\n", num_interactions);
	// end of handling, Diament & Tuller, May 2014
	if (use_weights == 5 || use_weights == 6)
		compute_wish_distances(interactions, use_weights);

	num_rDNA_interactions = 0;
	rDNA_interactions = NULL;
	if (strcmp(rDNA_interactions_filename, "") != 0) {
		num_rDNA_interactions = add_interactions(rDNA_interactions_filename,
				rDNA_interactions, chromosome, inter_priority);
	}

	num_unseen_interactions = 0;
	unseen_interactions = NULL;
	if (strcmp(unseen_interactions_filename, "") != 0) {
		num_unseen_interactions = add_interactions(unseen_interactions_filename,
				unseen_interactions, chromosome, inter_priority);
	}

}

/*****************************************************************************
 * Allocate a new yeast genome, haploid
 *****************************************************************************/
void GENOME::new_haploid_yeast_genome(int *chromosome) {
	int i;
	this->num_chroms = 0;
	for (i = 0; i < NUM_CHROMS; i++) {
		this->num_chroms += chromosome[i];
	}

	this->chroms = (CHROM_T**) mymalloc(sizeof(CHROM_T*) * this->num_chroms);
	int index = 0;
	for (i = 0; i < NUM_CHROMS; i++) {
		if (chromosome[i]) {
			char id[10];
			sprintf(id, "%d", i + 1);
			this->chroms[index] = new_chrom(id, 1, chrlen[i + 1]);
			index++;
		}
	}
	this->num_interactions = 0;
	this->interactions = NULL;
}

/*****************************************************************************
 * Allocate a new yeast genome, haploid
 *****************************************************************************/
void GENOME::new_haploid_yeast_genome() {

	this->num_chroms = NUM_CHROMS;
	this->chroms = (CHROM_T**) mymalloc(sizeof(CHROM_T*) * NUM_CHROMS);
	this->chroms[0] = new_chrom("1", 1, chrlen[1]);
	this->chroms[1] = new_chrom("2", 1, chrlen[2]);
	this->chroms[2] = new_chrom("3", 1, chrlen[3]);
	this->chroms[3] = new_chrom("4", 1, chrlen[4]);
	this->chroms[4] = new_chrom("5", 1, chrlen[5]);
	this->chroms[5] = new_chrom("6", 1, chrlen[6]);
	this->chroms[6] = new_chrom("7", 1, chrlen[7]);
	this->chroms[7] = new_chrom("8", 1, chrlen[8]);
	this->chroms[8] = new_chrom("9", 1, chrlen[9]);
	this->chroms[9] = new_chrom("10", 1, chrlen[10]);
	this->chroms[10] = new_chrom("11", 1, chrlen[11]);
	this->chroms[11] = new_chrom("12", 1, chrlen[12]);
	this->chroms[12] = new_chrom("13", 1, chrlen[13]);
	this->chroms[13] = new_chrom("14", 1, chrlen[14]);
	this->chroms[14] = new_chrom("15", 1, chrlen[15]);
	this->chroms[15] = new_chrom("16", 1, chrlen[16]);

	this->num_interactions = 0;
	this->interactions = NULL;

}

/*****************************************************************************
 * Allocate a new yeast genome, haploid
 *****************************************************************************/
void GENOME::new_haploid_yeast_genome(int chromosome) {
	this->num_chroms = 1;
	this->chroms = (CHROM_T**) mymalloc(sizeof(CHROM_T*) * NUM_CHROMS);
	char chrnum[5];
	sprintf(chrnum, "%d", chromosome);
	this->chroms[0] = new_chrom(chrnum, 1, chrlen[chromosome]);
	this->num_interactions = 0;
	this->interactions = NULL;

}

/*****************************************************************************
 * Allocate a haploid genome, made of only chromosome 4
 *****************************************************************************/
void GENOME::new_haploid_yeast_genome_chr4() {
	this->num_chroms = 1;
	this->chroms = (CHROM_T**) mymalloc(sizeof(CHROM_T*) * NUM_CHROMS);
	this->chroms[0] = new_chrom("4", 1, chrlen[4]);

	this->num_interactions = 0;
	this->interactions = NULL;

}

/*****************************************************************************
 * Allocate a haploid genome, made of only chromosome 3
 *****************************************************************************/
void GENOME::new_haploid_yeast_genome_chr3() {
	this->num_chroms = 1;
	this->chroms = (CHROM_T**) mymalloc(sizeof(CHROM_T*) * NUM_CHROMS);
	this->chroms[0] = new_chrom("3", 1, chrlen[3]);

	this->num_interactions = 0;
	this->interactions = NULL;

}

/*****************************************************************************
 * Allocate a haploid genome, made of only chromosome 3
 *****************************************************************************/
void GENOME::new_haploid_yeast_genome_chr34() {
	this->num_chroms = 2;
	this->chroms = (CHROM_T**) mymalloc(sizeof(CHROM_T*) * NUM_CHROMS);
	this->chroms[0] = new_chrom("3", 1, chrlen[3]);
	this->chroms[1] = new_chrom("4", 1, chrlen[4]);

	this->num_interactions = 0;
	this->interactions = NULL;

}

/*****************************************************************************
 * Compute the wish distances
 * To get the exact numbers, I used the file 2009-10-09/convert_frequencies_into_distances.m
 *****************************************************************************/

void GENOME::compute_wish_distances(INTERACTION_T* &interactions,
		int use_weights) {
	INTERACTION_T *interaction = interactions;
	// reporting, Diament & Tuller, May 2014
	if (use_weights == 5)
		printf("converting values to distances...\n");
	else
		printf("values are distances...\n");

	while (interaction != NULL) {
		int freq;
		if (use_weights == 5)   // the sum of frequencies
			freq = interaction->freq;
		else if (use_weights == 6)
			freq = (int) ((1.0 * interaction->freq) / interaction->count);
		else {
			printf("Use_weights should be 5 or 6 and is %d, ABORT!\n",
					use_weights);
			exit(1);
		}
		double wish_dist;

		//To get the exact numbers, I ran 2010-01-04/convert_frequencies_into_distances.m

		if (use_weights == 5) {
			if (freq >= 248)
				wish_dist = 19.230769;
			else if (freq >= 170 && freq <= 247)
				wish_dist = 38.221154;
			else if (freq >= 153 && freq <= 169)
				wish_dist = 57.211538;
			else if (freq >= 112 && freq <= 152)
				wish_dist = 76.201923;
			else if (freq >= 103 && freq <= 111)
				wish_dist = 95.192308;
			else if (freq >= 84 && freq <= 102)
				wish_dist = 114.182692;
			else if (freq >= 67 && freq <= 83)
				wish_dist = 133.173077;
			else if (freq >= 66 && freq <= 66)
				wish_dist = 152.163462;
			else if (freq >= 54 && freq <= 65)
				wish_dist = 171.153846;
			else if (freq >= 48 && freq <= 53)
				wish_dist = 190.144231;
			else if (freq >= 45 && freq <= 47)
				wish_dist = 209.134615;
			else if (freq >= 39 && freq <= 44)
				wish_dist = 228.125000;
			else if (freq >= 36 && freq <= 38)
				wish_dist = 247.115385;
			else if (freq >= 32 && freq <= 35)
				wish_dist = 266.105769;
			else if (freq >= 29 && freq <= 31)
				wish_dist = 285.096154;
			else if (freq >= 28 && freq <= 28)
				wish_dist = 304.086538;
			else if (freq >= 25 && freq <= 27)
				wish_dist = 323.076923;
			else if (freq >= 23 && freq <= 24)
				wish_dist = 342.067308;
			else if (freq >= 20 && freq <= 22)
				wish_dist = 380.048077;
			else if (freq >= 18 && freq <= 19)
				wish_dist = 418.028846;
			else if (freq >= 17 && freq <= 17)
				wish_dist = 437.019231;
			else if (freq >= 16 && freq <= 16)
				wish_dist = 456.009615;
			else if (freq >= 15 && freq <= 15)
				wish_dist = 475.000000;
			else if (freq >= 14 && freq <= 14)
				wish_dist = 493.990385;
			else if (freq >= 13 && freq <= 13)
				wish_dist = 531.971154;
			else if (freq >= 12 && freq <= 12)
				wish_dist = 550.961538;
			else if (freq >= 11 && freq <= 11)
				wish_dist = 588.942308;
			else if (freq >= 10 && freq <= 10)
				wish_dist = 626.923077;
			else if (freq >= 9 && freq <= 9)
				wish_dist = 683.894231;
			else if (freq >= 8 && freq <= 8)
				wish_dist = 740.865385;
			else if (freq >= 7 && freq <= 7)
				wish_dist = 835.817308;
			else if (freq >= 6 && freq <= 6)
				wish_dist = 949.759615;
			else if (freq >= 5 && freq <= 5)
				wish_dist = 1158.653846;
			else if (freq >= 4 && freq <= 4)
				wish_dist = 1272.596154;
			else if (freq >= 3 && freq <= 3)
				wish_dist = 1519.471154;
			else if (freq >= 1 && freq <= 2)
				wish_dist = 1538.461538;
		} else
			// use_weights=6, where we treat the input as already converted to distance
			// Diament & Tuller, May 2014
			wish_dist = freq;

		// those were nano-meters, so make sure I make them micro-meters
		interaction->wish_dist = wish_dist / 1000.0;
		//printf ("freq=%d, count=%d, avg=%d, wish dist = %lf\n", interaction->freq, interaction->count, freq, interaction->wish_dist);
		interaction = interaction->next;
	}
}

/*****************************************************************************
 * Allocate a new chromosome.
 *****************************************************************************/
int GENOME::add_interactions(char *interactions_filename,
		INTERACTION_T* &interactions, int *chromosome, int priority) {
	int num_int = 0;
	// Punt if no file was provided.
	if (interactions_filename == NULL) {
		return 0;
	}

	// Open the file for reading.
	FILE* interactions_file;
	if (open_file(interactions_filename, "r", TRUE, "interactions",
			"interactions", &interactions_file) == 0) {
		exit(1);
	}

	// interactions from the same input file can be merged
	// otherwise conflicts are discarded
	// Diament & Tuller, May 2014
	FILE* merge_log = fopen( "merge.log", "a");
	FILE* discard_log = fopen( "discard.log", "a");

	// Build the linked list of interactions.
	INTERACTION_T* this_interaction;
	INTERACTION_T* found_interaction;
	int line_number = 0;
	int count_merge = 0; // Diament & Tuller, May 2014
	int count_replace = 0; // Diament & Tuller, May 2014
	int count_discard = 0; // Diament & Tuller, May 2014
	while ((this_interaction = get_next_interaction(line_number + 1,
			interactions_file, chromosome)) != NULL)
	// if it's the same locus, get_next_interaction won't add it
	{
		//printf ("Processing line number %d\n", line_number);
		this_interaction->priority = priority; // Diament & Tuller, May 2014
		// Don't add an interaction that we already saw.
		// This is possible due to the reduced precision
		// When using intensities, maybe I should sum up the values
		found_interaction = has_interaction(this_interaction, interactions);
		if (found_interaction != NULL) {
			// priority is highest for 0 and reduces as numbers increase
			// Diament & Tuller, May 2014
			int prefer_this = found_interaction->priority
					- this_interaction->priority;
			if (prefer_this == 0) {
				//printf ("Interaction exists already, pref freq = %d!!\n", found_interaction->freq);
				count_merge += 1;
				fprintf( merge_log, "line: %d\texist\t%d\t%d\t%d\t%d\t%d\tnew\t%d\t%d\t%d\t%d\t%d\n",
						line_number,
						found_interaction->chrom1_index, found_interaction->locus1_index,
						found_interaction->chrom2_index, found_interaction->locus2_index,
						found_interaction->freq,
						this_interaction->chrom1_index, this_interaction->locus1_index,
						this_interaction->chrom2_index, this_interaction->locus2_index,
						this_interaction->freq);
				found_interaction->freq += this_interaction->freq;
				found_interaction->count += this_interaction->count; // will be used for averaging
				found_interaction->inversep += this_interaction->inversep;
				found_interaction->inverseq += this_interaction->inverseq;
				//printf ("  increasing frequency to %d\n", found_interaction->freq);
			} else if (prefer_this > 0) {
				// discard found, Diament & Tuller, May 2014
				//printf("replacing existing interaction\n", line_number);
				count_replace += 1;
				found_interaction->freq = this_interaction->freq;
				found_interaction->count = this_interaction->count;
				found_interaction->inversep = this_interaction->inversep;
				found_interaction->inverseq = this_interaction->inverseq;
				found_interaction->priority = this_interaction->priority;
				// it's probably better to leave the pointer of found_interaction
				// intact because there is a ->next pointing to it somewhere.
			} else if (prefer_this < 0) {
				//printf( "discarding interaction in line %d\n", line_number);
				/*printf( "existing: \nC_%d\tL_%d\tC_%d\tL_%d \ndiscarded: \nC_%d\tL_%d\tC_%d\tL_%d\n",
				 found_interaction->chrom1_index, found_interaction->locus1_index,
				 found_interaction->chrom2_index, found_interaction->locus2_index,
				 this_interaction->chrom1_index, this_interaction->locus1_index,
				 this_interaction->chrom2_index, this_interaction->locus2_index);*/
				count_discard += 1;
				fprintf( discard_log, "line: %d\texist\t%d\t%d\t%d\t%d\t%d\tnew\t%d\t%d\t%d\t%d\t%d\n",
						line_number,
						found_interaction->chrom1_index, found_interaction->locus1_index,
						found_interaction->chrom2_index, found_interaction->locus2_index,
						found_interaction->freq,
						this_interaction->chrom1_index, this_interaction->locus1_index,
						this_interaction->chrom2_index, this_interaction->locus2_index,
						this_interaction->freq);
				// discard this
			}
			myfree(this_interaction)
		} else {
			INTERACTION_T* next_interaction = interactions;
			interactions = this_interaction;
			this_interaction->next = next_interaction;
			num_int++;
		}
		line_number++;
	}
	printf("merged: %d, replaced: %d, discarded: %d\n", count_merge,
			count_replace, count_discard); // Diament & Tuller, May 2014
	fprintf(stdout, "Read %d interactions from %d lines in %s.\n", num_int,
			line_number, interactions_filename);

	fclose( merge_log); // Diament & Tuller, May 2014
	fclose( discard_log); // Diament & Tuller, May 2014
	fclose(interactions_file);
	return num_int;
}

/*****************************************************************************
 * Check whether a given interaction is in a linked list.
 * Mirela: Oct 17, 2009. If I find it, increase the frequency of the interaction that is already in the list
 * Mirela: Oct 22, 2009. Return a pointer to the interaction that has it, or nULL
 * Mirela: Oct 22, 2009. Don't increase the frequency here
 *****************************************************************************/
INTERACTION_T* GENOME::has_interaction(INTERACTION_T* query,
		INTERACTION_T* list) {
	if (list == NULL) {
		return (NULL);
	}

	if ((query->chrom1_index == list->chrom1_index)
			&& (query->locus1_index == list->locus1_index)
			&& (query->chrom2_index == list->chrom2_index)
			&& (query->locus2_index == list->locus2_index)) {
		return (list);
	}

	if ((query->chrom1_index == list->chrom2_index)
			&& (query->locus1_index == list->locus2_index)
			&& (query->chrom2_index == list->chrom1_index)
			&& (query->locus2_index == list->locus1_index)) {
		return (list);
	}

	return (has_interaction(query, list->next));
}

/*****************************************************************************
 * Read one interaction from a file.  Return the new interaction object.
 *****************************************************************************/
#define MAX_LINE 1000 // Maximum number of characters in one line.
INTERACTION_T* GENOME::get_next_interaction(int line_number,
		FILE* interactions_file, int *chromosome) {

	int chrom1;
	int chrom2;
	int locus1;
	int locus2;
	int dist;   // we are not interested in this, but it is in the file
	int freq;   // we are interested in frequency, and we'll use it as distance
	// TODO: the precision of float is not great, is up to e-38 to e38
	float pvalue;
	float qvalue;
	int chrom1_index;
	int chrom2_index;
	int locus1_index;
	int locus2_index;

	char one_line[MAX_LINE];

	while (1) {
		if (fgets(one_line, MAX_LINE, interactions_file) == NULL) {
			return (NULL);
		}
		//printf ("\n%s\n", one_line);

		// the new format has numbers for chromosomes, not codes
		//int num_scanned = sscanf(one_line, "%s %d %s %d",
		//                        chrom1_long, &locus1, chrom2_long, &locus2);
		int num_scanned = sscanf(one_line, "%d %d %d %d %d %d %e %e", &chrom1,
				&locus1, &chrom2, &locus2, &dist, &freq, &pvalue, &qvalue);

		if (num_scanned != 8) {
			die("Format error on line %d: <%s>\n", line_number, one_line);
		}

		// chrom1 and chrom2 start from 1
		// if chrom1 is wanted, chromosome[chrom1-1] should be 1
		if (chromosome[chrom1 - 1] && chromosome[chrom2 - 1]) {
			//printf ("Found the right chrom, break the loop\n");
			break;
		}

		// only keep the interactions for this chromosome
		//if (chrom1 != chromosome || chrom2 != chromosome)
		//{
		//printf ("Encountered chromosomes %d and %d, which are different from %d, IGNORE\n",
		//chrom1, chrom2, chromosome);
		// if I call this function recursively, I get stack overflow for big files
		//return get_next_interaction (line_number+1, interactions_file, chromosome);
		//}
	}

	// Convert chromosome long to chromosome id
//     chrom1_id = chrcode[string(chrom1_long)];
//     chrom2_id = chrcode[string(chrom2_long)];

	// do the rDNA adjustment
	if (add_rDNA) {
		if (chrom1 == 12)   // rDNA is on chrom 12
				{
			if (locus1 >= RDNA_POSITION)
				locus1 += SIZE_RDNA * NUM_RDNA_TO_ADD;
		}

		if (chrom2 == 12)   // rDNA is on chrom 12
				{
			if (locus2 >= RDNA_POSITION)
				locus2 += SIZE_RDNA * NUM_RDNA_TO_ADD;
		}
	}

	// Convert the chromosome ID to an index.
	chrom1_index = get_chrom_index(chrom1);   //chrom_id_to_index(chrom1_id);
	chrom2_index = get_chrom_index(chrom2);   //chrom_id_to_index(chrom2_id);

	// Reduce precision of locus coordinates.
	locus1_index = locus1 / bp_per_locus;
	locus2_index = locus2 / bp_per_locus;

//     printf ("Interaction chr %d %d, loci %d %d, read loci %d %d, bp-per-locus %d\n", 
//         chrom1_index, chrom2_index, locus1_index, locus2_index,
//         locus1, locus2, bp_per_locus);

	if (chrom1_index == chrom2_index && locus1_index == locus2_index) {
		// same locus, ignore
		//printf ("Same locus, ignore. %d==%d, %d==%d\n", chrom1_index, chrom2_index, locus1_index, locus2_index);
		return get_next_interaction(line_number + 1, interactions_file,
				chromosome);
	}

	INTERACTION_T* return_value = (INTERACTION_T*) mymalloc(
			sizeof(INTERACTION_T));

	return_value->chrom1_index = chrom1_index;
	return_value->chrom2_index = chrom2_index;
	return_value->locus1_index = locus1_index;
	return_value->locus2_index = locus2_index;
	return_value->freq = freq;

	if (pvalue < 1e-10)
		pvalue = 1e-10;
	if (qvalue < 1e-10)
		qvalue = 1e-10;
	return_value->inversep = 1.0 / pvalue;
	return_value->inverseq = 1.0 / qvalue;

	// tried to do a log transformation, but they are too close to each other
	//return_value->inversep = -1.0/logf(pvalue+FLT_MIN);
	//return_value->inverseq = -1.0/logf(qvalue+FLT_MIN);

	//printf ("%s\npvalue = %e, inversep = %lf, qvalue = %e, inverseq = %lf\n",
	//    one_line, pvalue, return_value->inversep, qvalue, return_value->inverseq);

	return_value->count = 1;

	//printf ("%d %d %d %d %d\n", chrom1, locus1, chrom2, locus2, freq);

	//printf ("Reading line:\n%s\n", one_line);
	//printf ("chr1: %s %s %d, locus1: %d %d\n", chrom1_long, chrom1_id.c_str(), return_value->chrom1_index, locus1, return_value->locus1_index);
	//printf ("chr2: %s %s %d, locus2: %d %d\n", chrom2_long, chrom2_id.c_str(), return_value->chrom2_index, locus2, return_value->locus2_index);

	// TODO: Not sure why/whether I need this, but it doesn't work well for inter-interactions
	/*
	 // Check validity by retrieving the corresponding coordinates.
	 if ((get_locus_coords(0, return_value->locus1_index,
	 get_nth_chrom(return_value->chrom1_index))
	 == NULL) ||
	 (get_locus_coords(0, return_value->locus2_index,
	 get_nth_chrom(return_value->chrom2_index))
	 == NULL)) {
	 die("Error on line %d: %s", line_number, one_line);
	 }
	 */

	return (return_value);
}

int GENOME::get_chrom_index(int chrom)
// chrom is 1-16
// return the index in our data structures. This depends on which chromosomes are wanted (chromosome)
		{
	int index = 0;
	for (int i = 1; i < chrom; i++) {
		if (wanted_chrom[i - 1])
			index++;
	}
	return index;
}

/*****************************************************************************
 * Retrieve a pointer to a specified locus, with bounds checking.
 *
 * Print an error and return NULL if illegal.
 *****************************************************************************/
COORDS_T* GENOME::get_locus_coords(int i_copy, int i_locus, CHROM_T* my_chrom) {
	if ((i_copy < 0) || (i_copy >= my_chrom->num_copies)) {
		fprintf(stderr,
				"Invalid copy index (%d) for chromosome %s with %d copies.\n",
				i_copy, my_chrom->id, my_chrom->num_copies);
		return (NULL);
	}
	if ((i_locus < 0) || (i_locus >= my_chrom->num_loci)) {
		fprintf(stderr,
				"Invalid locus index (%d) for chromosome %s with %d loci.\n",
				i_locus, my_chrom->id, my_chrom->num_loci);
		return (NULL);
	}
	return (my_chrom->loci[i_copy][i_locus]);
}

/*****************************************************************************
 * Retrieve a pointer to the nth chromosome.  Do bounds checking.
 *
 * Print an error and return NULL if illegal.
 *****************************************************************************/
CHROM_T* GENOME::get_nth_chrom(int chrom_index) {
	if ((chrom_index < 0) || (chrom_index >= num_chroms)) {
		fprintf(stderr, "Invalid chromosome index (%d).\n", chrom_index);
		return (NULL);
	}
	return (chroms[chrom_index]);
}

/*****************************************************************************
 * Read one interaction from a file.  Return the new interaction object.
 *****************************************************************************/
int GENOME::chrom_id_to_index(string chrom_id) {
	int return_value = atoi(chrom_id.c_str());
	cout << chrom_id;
	printf("Chrom id: %s %d\n", chrom_id.c_str(), return_value);
	if (return_value == 0) {
		if (chrom_id == "X") {
			return_value = 23;
		} else if (chrom_id == "Y") {
			return_value = 24;
		}
	}
	return_value -= 1;  // Indices are indexed from zero.
	if ((return_value < 0) || (return_value >= 24)) {
		die("Invalid chromosome ID (%s).\n", chrom_id.c_str());
	}
	return (return_value);
}

/*****************************************************************************
 * Allocate a new chromosome.
 *****************************************************************************/
CHROM_T* GENOME::new_chrom(char* id, int num_copies, int num_bases) {
	CHROM_T* return_value;

	//  fprintf(stderr, "Creating %d copies of chromosome %s.\n", num_copies, id);

	// Allocate the chromosome.
	return_value = (CHROM_T*) mymalloc(sizeof(CHROM_T));

	// Copy the chromosome ID.
	strncpy(return_value->id, id, 3);

	// Store the number of copies.
	return_value->num_copies = num_copies;

	// Compute and store the resolution, number of bases and number of loci.
	return_value->num_bases = num_bases;

	return_value->num_loci = (num_bases / bp_per_locus) + 1;
	return_value->loci = (COORDS_T***) mymalloc(
			sizeof(COORDS_T**) * num_copies);
	int i_copy = 0;
	for (i_copy = 0; i_copy < num_copies; i_copy++) {
		return_value->loci[i_copy] = (COORDS_T**) mymalloc(
				sizeof(COORDS_T*) * return_value->num_loci);

		// Put the initial locus at a random location.
		return_value->loci[i_copy][0] = new_coords();
		randomize_coords(return_value->loci[i_copy][0]);

		// Place subsequent loci near the previous one.
		int i_locus;
		for (i_locus = 1; i_locus < return_value->num_loci; i_locus++) {
			return_value->loci[i_copy][i_locus] = new_coords();
			pick_nearby_coords(max_locus_distance,
					return_value->loci[i_copy][i_locus - 1],
					return_value->loci[i_copy][i_locus]);
		}
	}

	return (return_value);
}

/*****************************************************************************
 * Print a genome in PDB format, to a specified output file
 *****************************************************************************/

void GENOME::print_pdb_genome(char *pdb_filename) {
	FILE* pdb_file;
	if (open_file(pdb_filename, "w", TRUE, "PDB", "PDB", &pdb_file) == 0) {
		exit(1);
	}
	print_pdb_genome(pdb_file);
	fclose(pdb_file);
}

/*****************************************************************************
 * Print a genome in PDB format, to a file already opened, or to stdout.
 *****************************************************************************/
#define PRINT_BEZIER 1
//#define PRINT_LINEAR 1

#ifdef PRINT_LINEAR
void GENOME::print_pdb_genome (FILE* outfile)
{
	int i_atom = 0;

	int i_chrom;
	for (i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];

		int i_copy;
		for (i_copy = 0; i_copy < this_chrom->num_copies; i_copy++) {

			int i_locus;
			for (i_locus = 0; i_locus < this_chrom->num_loci; i_locus++) {
				if (i_locus != 0) {

					print_pdb_edge(outfile, i_chrom, i_copy,
							this_chrom->loci[i_copy][i_locus-1],
							this_chrom->loci[i_copy][i_locus]);
					i_atom++;
				}

				print_pdb_coords(outfile, i_chrom, i_copy,
						this_chrom->loci[i_copy][i_locus]);
				i_atom++;
			}
		}
	}
}

void GENOME::print_1D_3D_genome (char *filename)
{
	FILE* outfile;
	if (open_file(filename, "w", TRUE, "1Dto3D", "1Dto3D", &outfile) == 0) {
		exit(1);
	}
	fprintf (outfile, "chrom\tlocus\t3D_x\t3D_y\t3D_z\n");

	int i_atom = 0;

	int i_chrom;
	for (i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];

		int i_copy;
		for (i_copy = 0; i_copy < this_chrom->num_copies; i_copy++) {

			int i_locus;
			for (i_locus = 0; i_locus < this_chrom->num_loci; i_locus++) {
				int from_1D = (i_locus-1)*bp_per_locus + bp_per_locus/2;
				int to_1D = (i_locus)*bp_per_locus + bp_per_locus/2;
				if (i_locus != 0) {
					print_1D_3D_edge(outfile, i_chrom, i_copy,
							this_chrom->loci[i_copy][i_locus-1],
							this_chrom->loci[i_copy][i_locus], from_1D, to_1D);
					i_atom++;
				}

				print_1D_3D_atom(outfile, i_chrom, i_copy,
						this_chrom->loci[i_copy][i_locus], i_locus*bp_per_locus+1);
				i_atom++;
			}
		}
	}
	fclose (outfile);
}
#endif

#ifdef PRINT_ELBOWS
#define ELBOW_SIZE 0.1
void GENOME::print_pdb_genome (FILE* outfile)
{
	int i_atom = 0;

	int i_chrom;
	for (i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];

		int i_copy;
		for (i_copy = 0; i_copy < this_chrom->num_copies; i_copy++) {

			int i_locus;
			for (i_locus = 0; i_locus < this_chrom->num_loci; i_locus++) {
				if (i_locus != 0) {
					COORDS_T p0,p1,p2;
					copy_coords(this_chrom->loci[i_copy][i_locus-1],&p1);
					copy_coords(this_chrom->loci[i_copy][i_locus],&p2);
					if(i_locus!=1) {
						copy_coords(this_chrom->loci[i_copy][i_locus-1],&p0);
						move_toward(&p1, this_chrom->loci[i_copy][i_locus], ELBOW_SIZE);
						move_toward(&p0,this_chrom->loci[i_copy][i_locus-2], ELBOW_SIZE);
						print_pdb_edge(outfile,i_chrom,i_copy,&p0,&p1);
					}
					if(i_locus!=this_chrom->num_loci-1)
					move_toward(&p2, this_chrom->loci[i_copy][i_locus-1], ELBOW_SIZE);
					print_pdb_edge(outfile, i_chrom, i_copy,&p1,&p2);
					//  this_chrom->loci[i_copy][i_locus-1],
					//  this_chrom->loci[i_copy][i_locus]);
					i_atom++;
				}

				// print_pdb_coords(outfile, i_chrom, i_copy,
				//                 this_chrom->loci[i_copy][i_locus]);
				i_atom++;
			}
		}
	}
}
#endif 

#ifdef PRINT_BEZIER
void GENOME::print_pdb_genome(FILE* outfile) {
	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];
		for (int i_copy = 0; i_copy < this_chrom->num_copies; i_copy++) {
			for (int i_locus = 0; i_locus < this_chrom->num_loci; i_locus++) {
				COORDS_T p0, p1, p2;
				if (i_locus == 0) {   // if it's the first locus
					copy_coords(this_chrom->loci[i_copy][0], &p0);
					copy_coords(this_chrom->loci[i_copy][1], &p1);
					move_toward_portion(&p1, &p0, 0.5); // move p1 to the middle between p0 and p1
					move_toward(&p0, &p1, 0.01); // move p0 a little bit towards the new (middle) p1
					print_pdb_edge(outfile, i_chrom, i_copy, &p0, &p1);
				} else if (i_locus == this_chrom->num_loci - 1) { // if it's the last locus
					copy_coords(
							this_chrom->loci[i_copy][this_chrom->num_loci - 2],
							&p0);
					copy_coords(
							this_chrom->loci[i_copy][this_chrom->num_loci - 1],
							&p1);
					move_toward_portion(&p0, &p1, 0.5);
					move_toward(&p1, &p0, 0.01);
					print_pdb_edge(outfile, i_chrom, i_copy, &p0, &p1);
				} else {      // a locus other than the first or last
					// compute Bezier control points
					copy_coords(this_chrom->loci[i_copy][i_locus], &p1);
					copy_coords(this_chrom->loci[i_copy][i_locus - 1], &p0);
					move_toward_portion(&p0, &p1, 0.5);
					copy_coords(this_chrom->loci[i_copy][i_locus + 1], &p2);
					move_toward_portion(&p2, &p1, 0.5);

					if (i_locus == centromeres[i_chrom] / bp_per_locus) {
						//printf ("Found centromere for chr %d\n", i_chrom+1);
						print_pdb_quadratic(outfile, i_chrom, i_copy, &p0, &p1,
								&p2, 1);
					} else
						print_pdb_quadratic(outfile, i_chrom, i_copy, &p0, &p1,
								&p2, 0);
				}

				if (i_locus == 0 || i_locus == this_chrom->num_loci - 1) {
					print_pdb_coords(outfile, i_chrom, i_copy,
							this_chrom->loci[i_copy][i_locus]);
				}
//         if (i_locus == centromeres[i_chrom]/bp_per_locus) {
//             print_pdb_atom(outfile, i_chrom, i_copy, FALSE, "C",
//                          this_chrom->loci[i_copy][i_locus]);
//         }

			}
		}
	}
}

void GENOME::print_1D_3D_genome(char *filename)
// this performs exactly the same steps as print_pdb_genome, except it doesn't print it in pdb format,
//  but in a tab-delimited format which has the 1D cordinates too
		{
	FILE* outfile;
	if (open_file(filename, "w", TRUE, "1Dto3D", "1Dto3D", &outfile) == 0) {
		exit(1);
	}
	fprintf(outfile, "chrom\tlocus\t3D_x\t3D_y\t3D_z\n");

	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];
		for (int i_copy = 0; i_copy < this_chrom->num_copies; i_copy++) {
			for (int i_locus = 0; i_locus < this_chrom->num_loci; i_locus++) {
				COORDS_T p0, p1, p2;
				/*            printf ("Chrom %d, locus %d, genpos %d - %d\n",
				 i_chrom+1, i_locus+1, i_locus*bp_per_locus+1, (i_locus+1)*bp_per_locus);
				 */
				if (i_locus == 0) {   // if it's the first locus
					// NOTE: this one should be here, not at the end of the if, to be printed before the first edge
					print_1D_3D_atom(outfile, i_chrom, i_copy,
							this_chrom->loci[i_copy][i_locus],
							i_locus * bp_per_locus + 1);

					copy_coords(this_chrom->loci[i_copy][0], &p0);
					copy_coords(this_chrom->loci[i_copy][1], &p1);
					move_toward_portion(&p1, &p0, 0.5); // move p1 to the middle between p0 and p1

					// do this before modifying p0
					int from_1D = i_locus * bp_per_locus + 1;
					int to_1D = i_locus * bp_per_locus + bp_per_locus / 2;
					move_toward_1D(from_1D, to_1D, &p0, &p1, 0.01);

					move_toward(&p0, &p1, 0.01); // move p0 a little bit towards the new (middle) p1
					// calculate the new genomic positions for p0 and p1
					print_1D_3D_edge(outfile, i_chrom, i_copy, &p0, &p1,
							from_1D, to_1D);
				} else if (i_locus == this_chrom->num_loci - 1) { // if it's the last locus
					copy_coords(
							this_chrom->loci[i_copy][this_chrom->num_loci - 2],
							&p0);
					copy_coords(
							this_chrom->loci[i_copy][this_chrom->num_loci - 1],
							&p1);
					move_toward_portion(&p0, &p1, 0.5);

					int from_1D = (i_locus - 1) * bp_per_locus
							+ bp_per_locus / 2;
					int to_1D = i_locus * bp_per_locus;
					move_toward_1D(from_1D, to_1D, &p1, &p0, 0.01);
					move_toward(&p1, &p0, 0.01);
					print_1D_3D_edge(outfile, i_chrom, i_copy, &p0, &p1,
							from_1D, to_1D);

					// print the last point (used to be after the if)

					print_1D_3D_atom(outfile, i_chrom, i_copy,
							this_chrom->loci[i_copy][i_locus],
							i_locus * bp_per_locus);
				} else {      // a locus other than the first or last
					// compute Bezier control points
					copy_coords(this_chrom->loci[i_copy][i_locus], &p1);
					copy_coords(this_chrom->loci[i_copy][i_locus - 1], &p0);
					int from_1D = (i_locus - 1) * bp_per_locus
							+ bp_per_locus / 2;
					int to_1D = (i_locus) * bp_per_locus + bp_per_locus / 2;

					move_toward_portion(&p0, &p1, 0.5);
					copy_coords(this_chrom->loci[i_copy][i_locus + 1], &p2);
					move_toward_portion(&p2, &p1, 0.5);

					print_1D_3D_quadratic(outfile, i_chrom, i_copy, &p0, &p1,
							&p2, from_1D, to_1D);

				}

			}
		}
	}

	fclose(outfile);
}

#endif

void GENOME::set_coords(double *x) {
	int i = 0;
	int i_chrom;
	for (i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];
		int i_copy;
		for (i_copy = 0; i_copy < this_chrom->num_copies; i_copy++) {
			int i_locus;
			for (i_locus = 0; i_locus < this_chrom->num_loci; i_locus++) {
				copy_coords_xyz(x[i++], x[i++], x[i++],
						this_chrom->loci[i_copy][i_locus]);
			}
		}
	}
}

int GENOME::get_x_index(int chrom_index, int locus_index) {
	// TODO: provide the chromosome copy too!
	//printf ("Chrom index: %d\n", chrom_index);
	CHROM_T* this_chrom = chroms[chrom_index];
	if (this_chrom->num_copies > 1) {
		printf("More than 1 chromosome copies, not implemented!\n");
		exit(1);
	}

	int index = 0;
	for (int i_chrom = 0; i_chrom < chrom_index; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];
		index += this_chrom->num_copies * this_chrom->num_loci;
	}
	index += locus_index;

	// multiply by 3 because there are 3 coordinates x, y and z
	index *= 3;

	return index;
}

/*****************************************************************************
 * Get the x, y and z coordinates
 * TODO: THis is stupidely slow, to redo, maybe use static
 *****************************************************************************/
COORDS_T * GENOME::get_coords(int index) {
	int index_sofar = 0;
	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];
		for (int i_copy = 0; i_copy < this_chrom->num_copies; i_copy++) {
			for (int i_locus = 0; i_locus < this_chrom->num_loci; i_locus++) {
				if (index == index_sofar)   // this is my locus
						{
					return this_chrom->loci[i_copy][i_locus];
				}
				index_sofar++;
			}
		}
	}
	return NULL;
}

/*****************************************************************************
 * 
 *****************************************************************************/
int GENOME::get_num_loci() {
	int num_loci = 0;
	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];
		num_loci += this_chrom->num_loci * this_chrom->num_copies;
	}
	return num_loci;
}

/*****************************************************************************
 * 
 *****************************************************************************/
int GENOME::get_chromosome(int locus) {
	int num_loci = 0;
	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];
		num_loci += this_chrom->num_loci * this_chrom->num_copies;
		if (locus <= num_loci)
			return i_chrom;
	}
	return -1;
}

/*****************************************************************************
 * 
 *****************************************************************************/
int GENOME::get_num_adjacent_loci() {
	int num_loci = 0;
	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		CHROM_T* this_chrom = chroms[i_chrom];
		num_loci += (this_chrom->num_loci - 1) * this_chrom->num_copies;
	}
	return num_loci;
}

/*****************************************************************************
 * Return the index that will be used in the binary matrix. 
 * The index will have loci for chrom 1, loci for chrom 2 etc 
 *****************************************************************************/
int GENOME::get_matrix_index(int chrom, int locus) {
	int index = 0;
	for (int j_chrom = 0; j_chrom < chrom; j_chrom++)
		index += chroms[j_chrom]->num_loci;
	index += locus;
	return index;
}

/*****************************************************************************
 * 
 *****************************************************************************/
void GENOME::save_adjacency_matrix_for_diffusion(char *output_filename,
		int operation_type)
// if operation_type is 0, print binary
// if operation_type is 1, print frequencies (sum of frequencies)
// if operation_type is 2, print counts
		{
	FILE* output_file;
	if (open_file(output_filename, "w", TRUE, "Interaction adjacency",
			"Interaction adjacency", &output_file) == 0) {
		exit(1);
	}
	// TODO: what happens if there are several copies of each chromosome?

	// firt print a row with headers
	fprintf(output_file, "Labels");
	for (int j_chrom = 0; j_chrom < num_chroms; j_chrom++) {
		for (int j_locus = 0; j_locus < chroms[j_chrom]->num_loci; j_locus++) {
			fprintf(output_file, "\tchr%d_locus%d", j_chrom + 1, j_locus + 1);
		}
//         if (j_chrom < num_chroms-1)
//             fprintf (output_file, "\t\t\t");  
	}
	fprintf(output_file, "\n");

	int matrix_size = 0;
	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		matrix_size += chroms[i_chrom]->num_loci;
	}

	// I'll store 1 and 0 in this matrix, so that I don't have to traverse the interactions linked list every time, but I just traverse it once
	int **matrix;
	matrix = new int*[matrix_size];
	for (int i = 0; i < matrix_size; i++) {
		matrix[i] = new int[matrix_size];
		for (int j = 0; j < matrix_size; j++)
			matrix[i][j] = 0;
	}
	INTERACTION_T *myint = interactions;
	while (myint != NULL) {
		int index1 = get_matrix_index(myint->chrom1_index, myint->locus1_index);
		int index2 = get_matrix_index(myint->chrom2_index, myint->locus2_index);
		//printf ("Chrom %d, locus %d, index %d\n", myint->chrom1_index, myint->locus1_index, index1);
		//printf ("Chrom %d, locus %d, index %d\n", myint->chrom2_index, myint->locus2_index, index2);
		if (operation_type == BINARY)    // binary
		{
			matrix[index1][index2] = 1;
			matrix[index2][index1] = 1;
		} else if (operation_type == FREQ) {
			matrix[index1][index2] = myint->freq;
			matrix[index2][index1] = myint->freq;
		} else if (operation_type == COUNT) {
			matrix[index1][index2] = myint->count;
			matrix[index2][index1] = myint->count;
		}
		myint = myint->next;
	}

	//printf ("Num chroms: %d\n", num_chroms);
	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		//printf ("Chrom %d, num loci %d\n", i_chrom, chroms[i_chrom]->num_loci);
		for (int i_locus = 0; i_locus < chroms[i_chrom]->num_loci; i_locus++) {
			fprintf(output_file, "chr%d_locus%d", i_chrom + 1, i_locus + 1);
			for (int j_chrom = 0; j_chrom < num_chroms; j_chrom++) {
				for (int j_locus = 0; j_locus < chroms[j_chrom]->num_loci;
						j_locus++) {
					int index1 = get_matrix_index(i_chrom, i_locus);
					int index2 = get_matrix_index(j_chrom, j_locus);

					// make sure the adjacent loci along the same chromosome have value 1
					// not sure I should do this, for now I'll only do it for binary
					if (operation_type == BINARY) {
						if (i_chrom == j_chrom
								&& (i_locus + 1 == j_locus
										|| i_locus == j_locus + 1))
							matrix[index1][index2] = 1;
					} else if (operation_type == COUNT) {
						if (i_chrom == j_chrom
								&& (i_locus + 1 == j_locus
										|| i_locus == j_locus + 1)
								&& matrix[index1][index2] == 0)
							matrix[index1][index2] = 1;
					}

					fprintf(output_file, "\t%d", matrix[index1][index2]);
				}
				/*                // finished this chromosome, add empty entries to show grey lines in the picture
				 if (j_chrom < num_chroms-1)
				 fprintf (output_file, "\t\t\t");  */
			}
			fprintf(output_file, "\n");
		}
		/*        // insert 3 empty lines
		 for (int k=1; k <= 3; k++)
		 {
		 if (i_chrom < num_chroms-1)
		 {
		 //fprintf (output_file, "empty");
		 for (int j_chrom = 0; j_chrom < num_chroms; j_chrom++)
		 {
		 for (int j_locus = 0; j_locus < chroms[j_chrom]->num_loci; j_locus++)
		 {
		 fprintf (output_file, "\t");
		 }
		 // finished this chromosome, add empty entries to show grey lines in the picture
		 if (j_chrom < num_chroms-1)
		 fprintf (output_file, "\t\t\t");
		 }
		 fprintf (output_file, "\n");
		 }
		 } */
	}
	fclose(output_file);
	for (int i = 0; i < matrix_size; i++)
		delete[] matrix[i];
	delete[] matrix;
}

/*****************************************************************************
 * 
 *****************************************************************************/
void GENOME::save_interaction_adjacency_matrix(char *output_filename) {
	FILE* output_file;
	if (open_file(output_filename, "w", TRUE, "Interaction adjacency",
			"Interaction adjacency", &output_file) == 0) {
		exit(1);
	}
	// TODO: what happens if there are several copies of each chromosome?

	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		for (int i_locus = 0; i_locus < chroms[i_chrom]->num_loci; i_locus++) {
			for (int j_chrom = 0; j_chrom < num_chroms; j_chrom++) {
				for (int j_locus = 0; j_locus < chroms[j_chrom]->num_loci;
						j_locus++) {
					// get the interaction score (or just 0/1 values)
					INTERACTION_T myint;
					myint.chrom1_index = i_chrom;
					myint.locus1_index = i_locus;
					myint.chrom2_index = j_chrom;
					myint.locus2_index = j_locus;
					if (has_interaction(&myint, interactions))
						fprintf(output_file, "1 ");
					else
						fprintf(output_file, "0 ");
				}
				if (j_chrom != num_chroms - 1)
					fprintf(output_file, "0.3 ");
			}
			fprintf(output_file, "\n");
		}
		// add a row of 0.5 between chromosomes, for visualization purposes
		if (i_chrom != num_chroms - 1) {
			for (int j_chrom = 0; j_chrom < num_chroms; j_chrom++)
				for (int j_locus = 0; j_locus < chroms[j_chrom]->num_loci;
						j_locus++)
					fprintf(output_file, "0.3 ");
			fprintf(output_file, "0.3\n");
		}
	}
	fclose(output_file);
}

/*****************************************************************************
 * 
 *****************************************************************************/

void GENOME::save_interaction_achievement(char *output_filename) {
	FILE* output_file;
	if (open_file(output_filename, "w", TRUE, "Interaction achievement",
			"Interaction achievement", &output_file) == 0) {
		exit(1);
	}
	// TODO: what happens if there are several copies of each chromosome?

	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		for (int i_locus = 0; i_locus < chroms[i_chrom]->num_loci; i_locus++) {
			for (int j_chrom = 0; j_chrom < num_chroms; j_chrom++) {
				for (int j_locus = 0; j_locus < chroms[j_chrom]->num_loci;
						j_locus++) {
					// get the interaction score (or just 0/1 values)
					INTERACTION_T myint;
					myint.chrom1_index = i_chrom;
					myint.locus1_index = i_locus;
					myint.chrom2_index = j_chrom;
					myint.locus2_index = j_locus;

					COORDS_T *coord1 = chroms[i_chrom]->loci[0][i_locus];
					COORDS_T *coord2 = chroms[j_chrom]->loci[0][j_locus];
					double dist = compute_distance(coord1, coord2);

					if (has_interaction(&myint, interactions)) {
						fprintf(output_file, "%.4lf ", dist);
					} else
						fprintf(output_file, "0.01 ");
				}
				if (j_chrom != num_chroms - 1)
					fprintf(output_file, "0.012 ");
			}
			fprintf(output_file, "\n");
		}
		// add a row of 0.5 between chromosomes, for visualization purposes
		if (i_chrom != num_chroms - 1) {
			for (int j_chrom = 0; j_chrom < num_chroms; j_chrom++)
				for (int j_locus = 0; j_locus < chroms[j_chrom]->num_loci;
						j_locus++)
					fprintf(output_file, "0.012 ");
			fprintf(output_file, "0.012\n");
		}
	}
	fclose(output_file);
}

/*****************************************************************************
 * 
 *****************************************************************************/
void GENOME::save_optimal_distances(char *output_filename) {
	FILE* output_file;
	if (open_file(output_filename, "w", TRUE, "Optimal distances",
			"Optimal distances", &output_file) == 0) {
		exit(1);
	}
	// TODO: what happens if there are several copies of each chromosome?

	for (int i_chrom = 0; i_chrom < num_chroms; i_chrom++) {
		for (int i_locus = 0; i_locus < chroms[i_chrom]->num_loci; i_locus++) {
			for (int j_chrom = 0; j_chrom < num_chroms; j_chrom++) {
				for (int j_locus = 0; j_locus < chroms[j_chrom]->num_loci;
						j_locus++) {
					COORDS_T *coord1 = chroms[i_chrom]->loci[0][i_locus];
					COORDS_T *coord2 = chroms[j_chrom]->loci[0][j_locus];

					double dist = compute_distance(coord1, coord2);
					// only display the closest interactions
					if (1 - dist >= 0.9)
						fprintf(output_file, "%.4lf ", 1 - dist);
					else
						fprintf(output_file, "0.8 ");
				}
				if (j_chrom != num_chroms - 1)
					fprintf(output_file, "0.87 ");
			}
			fprintf(output_file, "\n");
		}

		// add a row of 0.5 between chromosomes, for visualization purposes
		if (i_chrom != num_chroms - 1) {
			for (int j_chrom = 0; j_chrom < num_chroms; j_chrom++)
				for (int j_locus = 0; j_locus < chroms[j_chrom]->num_loci;
						j_locus++)
					fprintf(output_file, "0.87 ");
			fprintf(output_file, "0.87\n");
		}
	}
	fclose(output_file);
}

void GENOME::read_txt_input(char *txt_filename) {
	FILE* file = fopen(txt_filename, "r");
	if (file == 0)
		die("Unable to open text input file. \n");

	int num_variables = get_num_loci() * 3;
	double *x = new double[num_variables];

	for (int i = 0; i < num_variables; i++) {
		fscanf(file, "%le\n", &x[i]);
	}
	fclose(file);
	set_coords((double*) x);
	delete[] x;
}

void GENOME::save_txt(char *txt_filename) {
	FILE* file = fopen(txt_filename, "w");

	for (int i = 0; i < get_num_loci(); i++) {
		COORDS_T *coord = get_coords(i);
		if (coord == NULL)
			die("Something wrong in GENOME::get_coords, returned NULL\n");
		fprintf(file, "%e\n", coord->x);
		fprintf(file, "%e\n", coord->y);
		fprintf(file, "%e\n", coord->z);
	}
	fclose(file);
}

#define MAX_LINE 1000 // Maximum number of characters in one line.
void GENOME::read_cplex_output(char *cplex_filename) {
	FILE* file;
	// first traverse the file and get the maximum index. That will be num_variables
	if (open_file(cplex_filename, "r", TRUE, "cplex", "cplex", &file) == 0) {
		exit(1);
	}
	char one_line[MAX_LINE];
	int xindex;
	double value;
	int num_variables = 0;
	while (1) {
		if (fgets(one_line, MAX_LINE, file) == NULL) {
			break;
		}
		int num_scanned = sscanf(one_line, "%d", &xindex);

		//printf ("%s, num_scanned = %d\n", one_line, num_scanned);
		if (num_scanned == 1) {
			if (xindex > num_variables)
				num_variables = xindex;
		}
	}
	fclose(file);
	printf("Num variables: %d\n", num_variables);

	// Open the file for reading.
	double *x = new double[num_variables];
	if (open_file(cplex_filename, "r", TRUE, "cplex", "cplex", &file) == 0) {
		exit(1);
	}
	while (1) {
		if (fgets(one_line, MAX_LINE, file) == NULL) {
			break;
		}
		int num_scanned = sscanf(one_line, "%d %le", &xindex, &value);
		printf("%s, num_scanned = %d\n", one_line, num_scanned);
		if (num_scanned == 2) {
			printf("x%d has value %e\n", xindex, value);
			x[xindex] = value;
		}
	}
	set_coords((double*) x);
	delete[] x;
	fclose(file);
}

